/*
 GBxCart RW - Console Interface Flasher
 Version: 1.16
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 26/08/2017
 Last Modified: 2/03/2019
 
 This program allows you to write ROM to Flash Carts that are supported.
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

int main(int argc, char **argv) {
	
	printf("GBxCart RW Flasher v1.16 by insideGadgets\n");
	printf("########################################\n");
	
	// Check arguments
	if (argc >= 2) {
		read_config();
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i\n", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		
		if (gbxcartFirmwareVersion <= 8) {
			printf("Firmware R9 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		
		// Separate the file name
		char filenameOnly[100];
		char filename[254];
		strncpy(filename, argv[1], 253);
		
		char *token = strtok(filename, "\\");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "\\");
		}
		
		printf("\n--- Write ROM to Flash Cart ---\n");
		printf("Cart selected: ");
		
		uint32_t readBytes = 0;
		uint16_t sector = 0;
		long fileSize = 0;
		
		// Grab file size
		FILE *romFile = fopen(argv[1], "rb");
		if (romFile != NULL) {
			fseek(romFile, 0, SEEK_END);
			fileSize = ftell(romFile);
			fseek(romFile, 0, SEEK_SET);
		}
		else {
			printf("\n%s \nFile not found\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		printf("32 MByte (S29GL128) Gameboy Advance Flash Cart\n");
		printf("\nGoing to write to ROM (Flash cart) from %s\n", filenameOnly);
		
		// Check file size
		if (fileSize > 0x2000000) {
			fclose(romFile);
			printf("\n%s \nFile size is larger than the available Flash cart space of 32 MBytes\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		printf("\nWriting to ROM (Flash cart) from %s\n", filenameOnly);
		printf("[             25%%             50%%             75%%            100%%]\n[");
		
		// Read rom a tiny bit before writing
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		set_mode(GBA_READ_ROM);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop();
		
		// Set end address as file size
		endAddr = fileSize;
		
		// Write ROM
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		while (currAddr < endAddr) {
			if (currAddr % 0x10000 == 0) { // Erase next sector
				printf("Erasing sector at 0x%X\n", currAddr);
				gba_flash_write_address_byte(0xAAA, 0xA9);
				gba_flash_write_address_byte(0x555, 0x56);
				gba_flash_write_address_byte(0xAAA, 0x80);
				gba_flash_write_address_byte(0xAAA, 0xA9);
				gba_flash_write_address_byte(0x555, 0x56);
				gba_flash_write_address_byte((uint32_t) sector << 16, 0x30);
				sector++;
				
				// Wait for first 2 bytes to be 0xFF
				readBuffer[0] = 0;
				readBuffer[1] = 0;
				while (readBuffer[0] != 0xFF && readBuffer[1] != 0xFF) {
					set_number(currAddr / 2, SET_START_ADDRESS);
					set_mode(GBA_READ_ROM);
					
					com_read_bytes(READ_BUFFER, 64);
					com_read_stop(); // End read
					
					printf("0x%X,0x%X ", readBuffer[0], readBuffer[1]);
					
					delay_ms(50);
				}
				
				printf("\n");
				set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
			}
			
			printf("Address 0x%X\n", currAddr);
			
			com_write_bytes_from_file(GBA_FLASH_WRITE_256BYTE_SWAPPED_D0D1, romFile, 256);
			com_wait_for_ack();
			currAddr += 256;
			readBytes += 256;
			
			print_progress_percent(readBytes, (endAddr) / 64);
		}
	}
	
	printf("\n");
	
	return 0;
}